''' Tests of datafetcher return codes:
    check that arguments passed return expected codes

'''

import os
from pathlib import Path

import pytest
from datafetcher.datafetcher import DataFetcher
from datafetcher.return_codes import ReturnCode
from datafetcher.utilities import get_arg_parser, RetrievalMode

from .mock_data_fetcher import MockProdDataFetcher
# N.B. IJ doesn't recognize imported fixtures as being in use.
# don't let these imports (make_tempdir, settings) get disappeared.
from .test_df_function import launch_datafetcher, RUN_ALL, \
    MISSING_PROFILE, MISSING_SETTING
from .df_pytest_utils import TEST_PROFILE, get_test_capo_settings, make_tempdir, \
    capo_settings, settings


# pylint: disable=C0115, C0116, E0401, E1101, R0902, R0903, R0914, W0613, W0621, W1203

def test_launch_df(make_tempdir, settings):
    args = ['--product-locator', settings.test_data['product_locator'],
            '--output-dir', str(make_tempdir),
            '--profile', TEST_PROFILE,
            '--dry-run'
            ]
    return_code = launch_datafetcher(args, settings.capo_settings)
    assert return_code is not None


def test_launch_df_no_args(settings):
    try:
        return_code = launch_datafetcher(None, settings.capo_settings)
        assert return_code is not None
    except Exception as exc:
        pytest.fail(f'{exc}')


@pytest.mark.skipif(not RUN_ALL, reason='debug')
def test_no_args_prints_usage_and_croaks():
    ''' If the datafetcher is called without any arguments,
        it should print the "usage" statement, then exit

    '''

    with pytest.raises(TypeError):
        datafetcher = DataFetcher()
        assert datafetcher.usage == DataFetcher._build_usage_message()


# @pytest.mark.skipif(not RUN_ALL, reason='debug')
def test_omitted_capo_args_returns_expected_code(make_tempdir, settings):
    ''' Be sure DF dies with appropriate error message if called without Capo profile '''

    args = ['--product-locator', settings.test_data['product_locator'],
            '--output-dir', str(make_tempdir)]

    with pytest.raises(SystemExit) as exc:
        launch_datafetcher(args, settings.capo_settings)
    assert exc.value.code == 1


@pytest.mark.skipif(not RUN_ALL, reason='debug')
def test_omitted_capo_value_returns_expected_code(make_tempdir, settings):
    ''' Be sure DF dies with appropriate error message if Capo profile
        argument but no profile value

    '''

    args = ['--product-locator', "we're not going to get this far",
            '--output-dir', str(make_tempdir),
            '--profile']
    with pytest.raises(SystemExit) as exc:
        launch_datafetcher(args, settings.capo_settings)
    assert exc.value.code == 1


@pytest.mark.skipif(not RUN_ALL, reason='debug')
def test_invalid_capo_profile_returns_expected_code(make_tempdir, settings):
    ''' Be sure DF dies with appropriate error message if called without Capo profile '''

    args = ['--profile', 'whatevs',
            '--product-locator', "we'll never get this far",
            '--output-dir', str(make_tempdir),
            ]
    return_code = launch_datafetcher(args, settings.capo_settings)
    assert return_code == MISSING_PROFILE


def we_are_in_docker():
    ''' if we are in a Docker container, the SHELL env var will not be set '''

    for key, _ in os.environ.items():
        if key == 'SHELL':
            return False

    return True


@pytest.mark.skipif(not RUN_ALL, reason='debug')
def test_inaccessible_output_dir_returns_expected_code(settings, make_tempdir):
    try:
        umask = os.umask(0o000)
        tmpdir = Path(make_tempdir)
        tmpdir.chmod(0o666)

        altered_capo = get_test_capo_settings()
        altered_capo['download_dir'] = str(tmpdir)

        args = ['--product-locator', settings.test_data['product_locator'],
                '--profile', TEST_PROFILE,
                '--output-dir', str(tmpdir)]

        namespace = get_arg_parser().parse_args(args)

        # N.B. DataFetcher.__init__ throws SystemExit in pytest at command line,
        # but in Docker container the failure is interpreted as an Exception

        if we_are_in_docker():
            try:
                DataFetcher(namespace, capo_settings)

            except Exception as exc:
                assert isinstance(exc, SystemExit)
                assert exc.value.code == MISSING_SETTING
        else:
            with pytest.raises(SystemExit) as exc:
                DataFetcher(namespace, settings.capo_settings)
            assert exc.value.code == MISSING_SETTING

    except Exception as exc:
        pytest.fail(f'{exc}')
    finally:
        try:
            os.umask(umask)
        except Exception as exc:
            pytest.fail(f'{exc}')


@pytest.mark.skipif(not RUN_ALL, reason='debug')
def test_two_locator_args_returns_expected_code(make_tempdir, settings):
    ''' We should reject invocation with both product locator -and- location
        file. One or the other, people!

    '''

    args = ['--product-locator', 'a_locator',
            '--location-file', 'location.json'
                               '--output-dir', str(make_tempdir),
            '--profile', TEST_PROFILE]
    return_code = launch_datafetcher(args, settings.capo_settings)
    assert return_code == MISSING_SETTING


class MockServiceTimeoutReturn:
    ''' Simulates locator request service timeout '''

    @staticmethod
    def run():
        return ReturnCode.LOCATOR_SERVICE_TIMEOUT.value['code']


@pytest.mark.skipif(not RUN_ALL, reason='debug')
def test_locator_service_timeout_returns_expected_code(monkeypatch,
                                                       settings,
                                                       make_tempdir):
    def mock_run(*args, **kwargs):
        return MockServiceTimeoutReturn.run()

    args = ['--product-locator', settings.test_data['product_locator'],
            '--output-dir', str(make_tempdir),
            '--profile', TEST_PROFILE]
    monkeypatch.setattr(DataFetcher, "run", mock_run)
    return_code = launch_datafetcher(args, settings.capo_settings)
    assert return_code == ReturnCode.LOCATOR_SERVICE_TIMEOUT.value['code']


class MockTooManyServiceRedirectsReturn:

    @staticmethod
    def run():
        return ReturnCode.TOO_MANY_SERVICE_REDIRECTS.value


@pytest.mark.skipif(not RUN_ALL, reason='debug')
def test_too_many_service_redirects_returns_expected_code(monkeypatch,
                                                          settings,
                                                          make_tempdir):
    def mock_run(*args, **kwargs):
        return MockTooManyServiceRedirectsReturn().run()

    args = ['--product-locator', settings.test_data['product_locator'],
            '--output-dir', str(make_tempdir),
            '--profile', TEST_PROFILE]
    monkeypatch.setattr(DataFetcher, "run", mock_run)
    return_code = launch_datafetcher(args, settings.capo_settings)
    assert return_code == ReturnCode.TOO_MANY_SERVICE_REDIRECTS.value


class MockCatastrophicServiceErrorReturn:

    @staticmethod
    def run():
        return ReturnCode.CATASTROPHIC_REQUEST_ERROR


@pytest.mark.skipif(not RUN_ALL, reason='debug')
def test_catastrophic_service_error_returns_expected_code(monkeypatch,
                                                          settings,
                                                          make_tempdir):
    def mock_run(*args, **kwargs):
        return MockCatastrophicServiceErrorReturn().run()

    args = ['--product-locator', settings.test_data['product_locator'],
            '--output-dir', str(make_tempdir),
            '--profile', TEST_PROFILE]

    monkeypatch.setattr(DataFetcher, "run", mock_run)
    return_code = launch_datafetcher(args, settings.capo_settings)
    assert return_code == ReturnCode.CATASTROPHIC_REQUEST_ERROR


def test_copy_attempt_throws_sys_exit_service_error(monkeypatch,
                                                    settings,
                                                    make_tempdir):
    args = ['--product-locator', settings.test_data['product_locator'],
            '--output-dir', str(make_tempdir),
            '--profile', TEST_PROFILE]

    namespace = get_arg_parser().parse_args(args)
    fetcher = MockProdDataFetcher(namespace, settings.capo_settings)

    servers_report = fetcher.servers_report
    for server in servers_report:
        entry = servers_report[server]
        assert entry['retrieve_method'].value == RetrievalMode.COPY.value

    # let's try just one file so we're not sitting here all day
    for server in servers_report:
        entry = servers_report[server]
        servers_report = {server: entry}
        fetcher.servers_report = servers_report
        assert fetcher.servers_report[server] is not None
        break
    files = fetcher.servers_report[server]['files']
    fetcher.servers_report[server]['files'] = [files[0]]

    with pytest.raises(SystemExit) as exc:
        fetcher.run()
    assert exc.value.code == \
           ReturnCode.CATASTROPHIC_REQUEST_ERROR.value['code']


@pytest.mark.skipif(not RUN_ALL, reason='debug')
def test_product_locator_not_found_returns_expected_code(make_tempdir,
                                                         settings):
    args = ['--product-locator', 'not_a_locator',
            '--output-dir', str(make_tempdir),
            '--profile', TEST_PROFILE]
    return_code = launch_datafetcher(args, settings.capo_settings)
    assert return_code == ReturnCode.PRODUCT_LOCATOR_NOT_FOUND.value['code']


@pytest.mark.skipif(not RUN_ALL, reason='debug')
def test_unable_to_open_location_file_returns_expected_code(make_tempdir,
                                                            settings):
    args = ['--location-file', 'location.json',
            '--output-dir', str(make_tempdir),
            '--profile', TEST_PROFILE]
    return_code = launch_datafetcher(args, settings.capo_settings)
    assert return_code == ReturnCode.CANNOT_OPEN_LOCATION_FILE.value['code']


class MockNgasFetchError:
    ''' Simulates a problem fetching file(s) from NGAS server '''

    @staticmethod
    def run():
        return ReturnCode.NGAS_FETCH_ERROR


@pytest.mark.skipif(not RUN_ALL, reason='debug')
def test_error_fetching_file_from_ngas_returns_expected_code(monkeypatch,
                                                             settings,
                                                             make_tempdir):
    def mock_run(*args, **kwargs):
        return MockNgasFetchError().run()

    args = ['--product-locator', settings.test_data['product_locator'],
            '--output-dir', str(make_tempdir),
            '--profile', TEST_PROFILE]
    monkeypatch.setattr(DataFetcher, "run", mock_run)
    return_code = launch_datafetcher(args, settings.capo_settings)
    assert return_code == ReturnCode.NGAS_FETCH_ERROR


class MockSizeMismatchError:
    ''' Simulates a retrieved file of unexpected size '''

    @staticmethod
    def run():
        return ReturnCode.SIZE_MISMATCH


@pytest.mark.skipif(not RUN_ALL, reason='debug')
def test_unexpected_size_returns_expected_code(monkeypatch,
                                               settings,
                                               make_tempdir):
    def mock_run(*args, **kwargs):
        return MockSizeMismatchError().run()

    args = ['--product-locator', settings.test_data['product_locator'],
            '--output-dir', str(make_tempdir),
            '--profile', TEST_PROFILE]
    monkeypatch.setattr(DataFetcher, "run", mock_run)
    return_code = launch_datafetcher(args, settings.capo_settings)
    assert return_code == ReturnCode.SIZE_MISMATCH
